//-----------------------------------------------------------------------
//  RSARSim.cs 
//  Created by Bennie Lewis 
//  University of Central Florida 
//  Intelligent Agents Lab
//  Description: establish the Simulated level and creates all entities for the environment.
//  Name:            Date:                  Update/Change/Fix:
//  Bennie Lewis     02/03/2009             Created file
//-----------------------------------------------------------------------
using System;
using System.Collections.Generic;
using System.ComponentModel;
using Microsoft.Ccr.Core;
using Microsoft.Dss.Core.Attributes;
using Microsoft.Dss.ServiceModel.Dssp;
using Microsoft.Dss.ServiceModel.DsspServiceBase;
using W3C.Soap;
using submgr = Microsoft.Dss.Services.SubscriptionManager;

using Microsoft.Robotics.PhysicalModel;
using Microsoft.Robotics.Simulation;
using Microsoft.Robotics.Simulation.Engine;
using engineproxy = Microsoft.Robotics.Simulation.Engine.Proxy;
using Microsoft.Robotics.Simulation.Physics;

//---Simulated Services for Pioneer 3DX Robot----------------------------------------
using drive = Microsoft.Robotics.Services.Simulation.Drive.Proxy;
using lrf = Microsoft.Robotics.Services.Simulation.Sensors.LaserRangeFinder.Proxy;
using bumper = Microsoft.Robotics.Services.Simulation.Sensors.Bumper.Proxy;
using simwebcam = Microsoft.Robotics.Services.Simulation.Sensors.SimulatedWebcam.Proxy;
//------------------------------------------------------------------------------------

namespace RSARSim
{
    [Contract(Contract.Identifier)]
    [DisplayName("RSARSim")]
    [Description("RSARSim service (no description provided)")]
    class RSARSimService : DsspServiceBase
    {

        // Service state
        [ServiceState]
        RSARSimState _state = new RSARSimState();

        // partner attribute will cause simulation engine service to start
        [Partner("Engine",
            Contract = engineproxy.Contract.Identifier,
            CreationPolicy = PartnerCreationPolicy.UseExistingOrCreate)]
        private engineproxy.SimulationEnginePort _engineServicePort =
            new engineproxy.SimulationEnginePort();

 
        //Main service port
        [ServicePort("/RSARSim", AllowMultipleInstances = true)]
        RSARSimOperations _mainPort = new RSARSimOperations();

        [SubscriptionManagerPartner]
        submgr.SubscriptionManagerPort _submgrPort = new submgr.SubscriptionManagerPort();

        //Service constructor
        public RSARSimService(DsspServiceCreationPort creationPort)
            : base(creationPort)
        {
        }

        //Service start
        protected override void Start()
        {

            // 
            // Add service specific initialization here
            // 

            base.Start();
            // Orient sim camera view point
            SetupCamera();
            // Add objects (entities) in our simulated world
            PopulateWorld();
        }

        //Handles Subscribe messages
        /// <param name="subscribe">the subscribe request</param>
        [ServiceHandler]
        public void SubscribeHandler(Subscribe subscribe)
        {
            SubscribeHelper(_submgrPort, subscribe.Body, subscribe.ResponsePort);
        }

        private void SetupCamera()
        {
            // Set up initial view
            CameraView view = new CameraView();
            //view.EyePosition = new Vector3(2.491269f, 5.598689f, 1.046625f);
            //view.LookAtPoint = new Vector3(1.873792f, 5.40983f, 0.2830455f);
            view.EyePosition = new Vector3(10.0f, 30.0f, 5.5f);
            view.LookAtPoint = new Vector3(10.0f, 29.0f, 4.9f);
            SimulationEngine.GlobalInstancePort.Update(view);
        }

        private void PopulateWorld()
        {

            RescueLevel RL = new RescueLevel();
            RL.AddSky();
            RL.AddGround();
            RL.AddbrickLevel();
            RL.Addvictims();
            AddPioneer3DXRobot(new Vector3(-8.5f, 0.0f, -1.5f));
           
        }

        #region Pioneer3DX Robot 1
        //--------------------------Robot Entities--------------------------
        void AddPioneer3DXRobot(Vector3 position)
        {
            Pioneer3DX robotBaseEntity = CreateMotorBase(ref position);

            // Create Laser entity and start simulated laser service
            LaserRangeFinderEntity laser = CreateLaserRangeFinder();
            // insert laser as child to motor base
            robotBaseEntity.InsertEntity(laser);

            // Create bumper array entity and start simulated bumper service
            BumperArrayEntity bumperArray = CreateBumperArray();
            // insert as child of motor base
            robotBaseEntity.InsertEntity(bumperArray);

            // create Camera Entity ans start SimulatedWebcam service
            CameraEntity camera = CreateCamera();
            // insert as child of motor base
            robotBaseEntity.InsertEntity(camera);

            // if you want a image right above the robot uncomment"
            /*CameraSprite camSprite = new CameraSprite(0.5f, 0.5f, SpritePivotType.Center, 0, new Vector3(0, 1, 0));
            camSprite.State.Name = GetStringWithUniquePostfix("camSprite");
            camSprite.Flags |= VisualEntityProperties.DisableBackfaceCulling;
            camera.InsertEntity(camSprite);*/

            // insert the motor base and its two children 
            // to the simulation
            SimulationEngine.GlobalInstancePort.Insert(robotBaseEntity);
        }

        private Pioneer3DX CreateMotorBase(ref Vector3 position)
        {
            // use supplied entity that creates a motor base 
            // with 2 active wheels and one caster
            Pioneer3DX robotBaseEntity = new Pioneer3DX(position);

            // specify mesh. 
            robotBaseEntity.State.Assets.Mesh = "Pioneer3dx.bos";
            robotBaseEntity.WheelMesh = "PioneerWheel.bos";
            // specify color if no mesh is specified. 
            robotBaseEntity.ChassisShape.State.DiffuseColor = new Vector4(0.8f, 0.25f, 0.25f, 1.0f);

            // the name below must match manifest
            robotBaseEntity.State.Name = "P3DXMotorBase";

            // Start simulated arcos motor service
            drive.Contract.CreateService(ConstructorPort, "http://localhost/" + robotBaseEntity.State.Name,
                Microsoft.Robotics.Simulation.Partners.CreateEntityPartner(
                    "http://localhost/" + robotBaseEntity.State.Name)
            );

            return robotBaseEntity;
        }

        private LaserRangeFinderEntity CreateLaserRangeFinder()
        {
            // Create a Laser Range Finder Entity .
            // Place it 30cm above base CenterofMass. 
            LaserRangeFinderEntity laser = new LaserRangeFinderEntity(
                new Pose(new Vector3(0, 0.30f, 0)));

            laser.State.Name = "P3DXLaserRangeFinder";
            laser.LaserBox.State.DiffuseColor = new Vector4(0.25f, 0.25f, 0.8f, 1.0f);

            // Create LaserRangeFinder simulation service and specify
            // which entity it talks to
            lrf.Contract.CreateService(
                ConstructorPort, "http://localhost/" + laser.State.Name,
                Microsoft.Robotics.Simulation.Partners.CreateEntityPartner(
                "http://localhost/" + laser.State.Name));
            return laser;
        }

        private BumperArrayEntity CreateBumperArray()
        {
            // Create a bumper array entity with two bumpers
            BoxShape frontBumper = new BoxShape(
                new BoxShapeProperties("front",
                    0.001f,
                    new Pose(new Vector3(0, 0.05f, -0.25f)),
                    new Vector3(0.40f, 0.03f, 0.03f)
                )
            );
            frontBumper.State.DiffuseColor = new Vector4(0.1f, 0.1f, 0.1f, 1.0f);

            BoxShape rearBumper = new BoxShape(
                new BoxShapeProperties("rear",
                    0.001f,
                    new Pose(new Vector3(0, 0.05f, 0.25f)),
                    new Vector3(0.40f, 0.03f, 0.03f)
                )
            );
            rearBumper.State.DiffuseColor = new Vector4(0.1f, 0.1f, 0.1f, 1.0f);

            // The physics engine will issue contact notifications only
            // if we enable them per shape
            frontBumper.State.EnableContactNotifications = true;
            rearBumper.State.EnableContactNotifications = true;

            // put some force filtering so we only get notified for significant bumps
            //frontBumper.State.ContactFilter = new ContactNotificationFilter(1,1);
            //rearBumper.State.ContactFilter = new ContactNotificationFilter(1, 1);

            BumperArrayEntity
                bumperArray = new BumperArrayEntity(frontBumper, rearBumper);
            // entity name, must match manifest partner name
            bumperArray.State.Name = "P3DXBumpers";

            // start simulated bumper service

            bumper.Contract.CreateService(
                ConstructorPort, "http://localhost/" + bumperArray.State.Name,
                Microsoft.Robotics.Simulation.Partners.CreateEntityPartner(
                "http://localhost/" + bumperArray.State.Name));
            return bumperArray;
        }

        private CameraEntity CreateCamera()
        {
            // low resolution, wide Field of View
            CameraEntity cam = new CameraEntity(320, 240, CameraEntity.CameraModelType.AttachedChild);
            cam.State.Name = "Robot1_Cam";
            // just on top of the bot
            cam.State.Pose.Position = new Vector3(0.0f, 0.5f, 0.0f);
            // camera renders in an offline buffer at each frame
            // required for service
            cam.IsRealTimeCamera = true;

            // Start simulated webcam service
            simwebcam.Contract.CreateService(
                ConstructorPort, "http://localhost/" + cam.State.Name,
                Microsoft.Robotics.Simulation.Partners.CreateEntityPartner(
                    "http://localhost/" + cam.State.Name)
            );

            return cam;
        }

        //-------------------------Robot------------------------------------------------------
        #endregion
    }
}


